/* Copyright Airship and Contributors */

#import <Foundation/Foundation.h>
#import "UAAnalytics+Internal.h"
#import "UARequestSession.h"
#import "UAHTTPResponse.h"

@class UARuntimeConfig;

NS_ASSUME_NONNULL_BEGIN

/**
 * Event API response.
 */
@interface UAEventAPIResponse : UAHTTPResponse

/**
 * Maximum total database size in kB, as determined by the analytics server.
 */
@property(nonatomic, readonly, nullable) NSNumber *maxTotalDBSize;

/**
 * Maximum batch size in kB, as determined by the analytics server.
 */
@property(nonatomic, readonly, nullable) NSNumber *maxBatchSize;

/**
 * Minimum batch interval in seconds, as determined by the analytics server.
 */
@property(nonatomic, readonly, nullable) NSNumber *minBatchInterval;

/**
 * Init method.
 *
 * @param status The status code.
 * @param maxTotalDBSize The maximum total database size.
 * @param maxBatchSize The maximum batch size.
 * @param minBatchInterval The minimum batch interval.
 */
- (instancetype)initWithStatus:(NSUInteger)status
                maxTotalDBSize:(nullable NSNumber *)maxTotalDBSize
                  maxBatchSize:(nullable NSNumber *)maxBatchSize
              minBatchInterval:(nullable NSNumber *)minBatchInterval;

@end

/**
 * Represents possible event API client errors.
 */
typedef NS_ENUM(NSInteger, UAEventAPIClientError) {
    /**
     * Indicates an unsuccessful client status.
     */
    UAEventAPIClientErrorUnsuccessfulStatus
};

/**
 * The domain for NSErrors generated by the event API client.
 */
extern NSString * const UAEventAPIClientErrorDomain;

/**
 * API client to upload events to Airship.
 */
@interface UAEventAPIClient : NSObject

///---------------------------------------------------------------------------------------
/// @name Event API Client Internal Methods
///---------------------------------------------------------------------------------------

/**
 * Default factory method.
 *
 * @param config The Airship config.
 * @return A UAEventAPIClient instance.
 */
+ (instancetype)clientWithConfig:(UARuntimeConfig *)config;

/**
 * Factory method to create a UAEventAPIClient.
 *
 * @param config The Airship config.
 * @param session The UARequestSession instance.
 * @return UAEventAPIClient instance.
 */
+ (instancetype)clientWithConfig:(UARuntimeConfig *)config session:(UARequestSession *)session;

/**
 * Uploads analytic events.
 * @param events The events to upload.
 * @param headers The event headers.
 * @param completionHandler A completion handler.
 */
- (UADisposable *)uploadEvents:(NSArray *)events headers:(NSDictionary<NSString *, NSString *> *)headers completionHandler:(void (^)(UAEventAPIResponse * _Nullable, NSError * _Nullable))completionHandler;

@end

NS_ASSUME_NONNULL_END
